# Demo for key byte recovery against hardware Spook
Next, we detail an byte recovery attack against the 2 shares hardware implementation of Spook. To do so, we make use of [MCP-DPA](https://eprint.iacr.org/2014/409.pdf) on the output of the Sbox.

## Sensible value description
Clyde is a bitslice cipher. Its state is composed of 4 32-bit words. The Sbox is on 4-bit and takes as input one bit in each of the words. This first operations performed during a Clyde execution are the following:
```
// key initialization
k = [0xdeadbe01, 0xffffff02, tl0x01020303, 0x00000000]

// plaintext initialization
p = [0xdefeca7e, 0xdefeca7e, 0x11111111, 0x12345678]

// Key addition
x = k ^ p

// Sbox layer
y[1] = (x[0] & x[1]) ^ x[2]
y[0] = (x[3] & x[0]) ^ x[1]
y[3] = (y[1] & x[3]) ^ x[0]
y[2] = (y[0] & y[1]) ^ x[3]

```
In this demonstration, we target 8-bit of key which are at the input of two Sboxes. Therefore, we will recover the 2 LSB in each of the key words. In the previous example, 
the targeted key is so [0b01, 0b10, 0b11, 0b00]. The sensible value to target are the similar bits in y. This is next denoted as the sensible value s.
(For simplicity, we omit here the addition of the tweak key which is a zero string in the first Clyde call within Spook.)

## Attack based on the sensible value

### Profiling
In order to recover a key byte, one has first to profile the variance of the sensible value. This is performed by the script [profile_mcp_dpa.py](profile_mcp_dpa.py) which estimates the variance for each of the possible value of s. It also provides an estimation of the second order SNR on the sensible value and plot it. On that plot, one can observe first the SNR and then the averaged power consumption (just the 3 first cycles are used here).

![Profiling Output](.figs/profile.png)

### Attack
Afterwards, these profiles can be used to perform the MCP-DPA. Namely, for each key candidates, the sensible value is guessed and the actual variance of the leakage is correlated to the expected variance. This is performed by the [attack_mcp_dpa.py](attack_mcp_dpa.py) script
The correlation for each of the key candidates can be plotted. In the next figure, the one of the correct key in red.
![Correlation](.figs/attack.png)

## How to probe internal states

Next, we detail how to probe internal states within the Spook architecture during a profiling phase. To do so, we make use of [verilator](https://www.veripool.org/wiki/verilator) . This one allows to build a compiled version of the HDL code to enable fast simulations. 

### Binary from HDL source code

Once Verilator is [installed](https://www.veripool.org/projects/verilator/wiki/Installing) on the system, one can compile the circuit by running:
```
make LIBSIM=libsim.so D=D DIRHDL=DIRHDL
```
This will compile the architecture located at in DIRHDL dicteroy with D shares. It outputs a shared library call libsim.so with the hardware commands specified. This is done in [tb.cpp](tb.cpp).

## Get the register value

Once the shared library is compiled, one can use a ProbCircuit object to recover the register value given some input commands. This object call the shared library and feeds it with the given commands (ram\_in). The following example probes the state register for the 100 first cycles and then unmasks it to retrieve at which cycle the input of the sbox is manipulated.

```

# define cycles to probes 
cycles = np.arange(100).astype(np.uint32)

# initialize the probing object
probecircuit = ProbeCircuit(D=D,cycles=cycles,LIBSIM=args.simlib)

# obtain the register value at every probed cycles
reg_value = probecircuit.sim_run_probe(ram_in).reshape((len(cycles),D*4))

# change the reprensentation of the shares within the reg_value to fit the Python one at cycles
reg_value = rev_wire(reg_value,D=D)

# get the unmasked value of the register
reg_value_um = sharing_inputs(reg_value,D=D)

# get the cycle of interest for sbox_input
cycles_interest,_ = np.where(sbox_input[0] == reg_value_um)


```
The full example is run in [profile_mcp_dpa.py](profile_mcp_dpa.py) (even if not used for a MCP_DPA).
