# MIT License
#
# Copyright (c) 2020 UCLouvain
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import os
from interface.spook import sbox_layer
from stella_zero.estimator.mcp_dpa import MCP_DPA
import numpy as np
import argparse
from tqdm import tqdm
import pickle
#########################
#########################
# Parsing args
#########################
parser = argparse.ArgumentParser(description="Attack phase of MCP-DPA",
        formatter_class=argparse.RawTextHelpFormatter
        )
parser.add_argument(
        "-t", "--tname", default="templates", help="file name where templates are saved"
)
parser.add_argument(
    "-f", "--fname", default="traces", help="directory containing all traces"
)
parser.add_argument(
        "-d", "--dshares", default=2, type=int, help="number of shares"
        )
parser.add_argument(
        "-i", "--begin", default=0, type=int, help="initial index for the traces"
        )
parser.add_argument(
        "-e", "--end", default=4000, type=int, help="final index for the traces"
        )

args = parser.parse_args()
D = args.dshares

# load the template
fname = args.tname+'.npz'
dic = np.load(fname)
SM = dic["SM"]
u = dic["u"]
s = dic["s"]

# init mcp_dpa object from stella_zero
mcp_dpa = MCP_DPA(SM=SM[0],u=u[0],s=s[0],D=D,Ng=256)

print("# -----------")
print("# ----attack")
print("# -----------")
for f in os.listdir(args.fname):
    ###########"#####
    ## LOADING TRACES
    fname = os.path.join(args.fname, f)
    print("# file: %s"%(fname),flush=True)
    dic = np.load(fname, allow_pickle=True)
    traces = dic["traces"][:,args.begin:args.end]
    nonce = np.frombuffer(np.array(dic["nonce"]).tobytes(),dtype=np.uint32).reshape((len(traces[:,0]),4)).T

    # computing the guessed sbox output
    guess = np.zeros((256,len(traces)),dtype=np.uint32)
    for g in range(256):
        #putting the key guess in a bitlice fashion
        guess_state = np.zeros(4,dtype=np.uint32)
        guess_state[0] = (g>>0) & 0x3
        guess_state[1] = (g>>2) & 0x3
        guess_state[2] = (g>>4) & 0x3
        guess_state[3] = (g>>6) & 0x3

        # get sbox output
        guess_state = np.bitwise_xor(nonce.T,guess_state)
        guess_state = guess_state.T
        guess_state = np.array(sbox_layer(guess_state))

        # transform it into an 8-bit word
        guess[g,:] = 0
        guess[g,:] |= (guess_state[0]&0x3) << 0
        guess[g,:] |= (guess_state[1]&0x3) << 2
        guess[g,:] |= (guess_state[2]&0x3) << 4
        guess[g,:] |= (guess_state[3]&0x3) << 6

    # update the mcp_dpa
    mcp_dpa.fit(traces[:,:],guess[:,:].astype(np.uint16))

# Turning the correlation input log proba
# This is an trick, a heuristic
corr =  np.max(mcp_dpa._corr,axis=1)
corr = corr**2
corr /= np.sum(corr)
lcorr = np.log10(corr)

# Format the masks for the target key bits.
mask = (0x3 << 0) |(0x3 << 32) | (0x3 << 64) | (0x3 << 96)
masks = [(mask<<i) for i in range(0,32,2)]
scores  =[np.zeros(256) for i in range(0,32,2)]
scores[0] = lcorr
for (mask, score) in zip(masks, scores):
    print('{:x}:{}'.format(mask, ','.join(f'{x:.04f}' for x in score)))
