# MIT License
#
# Copyright (c) 2020 UCLouvain
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np
from tqdm import tqdm
import ctypes

def sharing_inputs(inputs,D=2):
    if inputs.ndim == 1:
        inputs = inputs.reshape((1,len(inputs)))
    state = np.zeros((len(inputs),4),dtype=np.uint32)
    state = inputs[:,:4].copy()
    for i in range(4):
        for d in range(1,D):
            state[:,i] = np.bitwise_xor(state[:,i],inputs[:,(4*d)+i])
    return state

def rev_wire(inputs,D=2):
    """
        This function is used to permute bits from sharing_bundle_state
        to match the standard sharing representation.

        see utils.py: sharing_inputs
    """
    outputs = np.zeros(inputs.shape,dtype=np.uint32)
    for b in range(len(inputs[0,:])*32):
        bit = (inputs[:,b//32] >> (b%32)) & 0x01;
        outputs[:,((b%D)*4) + (b//(32*D))] |= (bit) << ((b//D)%32);
    return outputs


class ProbeCircuit():
    """
        This object is used to recover state within the Spook architecture. 
        By default, the probed register is sharing_bungle_state. This one
        contains the state of the Clyde encryption.
        The probed register is composed of 4*D 32-bit words where D is the 
        number of shares. 
    """
    @staticmethod
    def wrap_function(lib, funcname, restype, argtypes):
        """Simplify wrapping ctypes functions"""
        func = lib.__getattr__(funcname)
        func.restype = restype
        func.argtypes = argtypes
        return func

    def __init__(self,D,cycles,LIBSIM="./libtb.so"):
        """
            Init the simulation module.

            D: number of shares used by the module
            cycle: is a uint32_t array. It contains the cycles at which the
                register must be probed. 
            LIBSIM: path the output of verilator library.
        """
        self._cycles = cycles.astype(np.uint32)
        self._ncycles = len(cycles)
        self._D = D
        self._lib = ctypes.CDLL(LIBSIM)

        self._sim_run_probe = self.wrap_function(self._lib,"sim_run_probe",None,[ctypes.POINTER(ctypes.c_uint32),
                                    ctypes.c_uint32,
                                    ctypes.POINTER(ctypes.c_uint32),
                                    ctypes.POINTER(ctypes.c_uint32),
                                    ctypes.c_uint32])

    def sim_run_probe(self,inputs,probes=None):
        """
            inputs: is a uint32_t numpy array corresponding to the inputs 
            commands sent to the Spook module. 
        """
        if probes is None:
            probes = np.zeros(len(self._cycles)*self._D *4,dtype=np.uint32)

        self._sim_run_probe(inputs.astype(np.uint32).ctypes.data_as(ctypes.POINTER(ctypes.c_uint32)),
            ctypes.c_uint32(len(inputs)),
            self._cycles.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32)),
            probes.ctypes.data_as(ctypes.POINTER(ctypes.c_uint32)),
            ctypes.c_uint32(self._ncycles))
        return probes
