# MIT License
#
# Copyright (c) 2020 UCLouvain
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from interface.spook import sbox_layer
import matplotlib.pyplot as plt
from stella_zero.evaluation.snr import SNROrder
from stella_zero.estimator.mcp_dpa import MCP_DPA
import numpy as np
import argparse
from tqdm import tqdm
from probecircuit import *
#########################
#########################
# Parsing args
#########################
parser = argparse.ArgumentParser(description="Performs the profiling for MCP-DPA",
        formatter_class=argparse.RawTextHelpFormatter
        )
parser.add_argument(
    "-f", "--fname", default="traces", help="file name where traces are saved"
)
parser.add_argument(
        "-s", "--simlib", default="libtb.so", help="compiled lib for probecircuit"
)
parser.add_argument(
        "-o", "--outname", default="templates_mcp_dpa", help="file name where templates are saved"
)
parser.add_argument(
    "-N", "--Nfiles", default=1, type=int, help="amount of files in which the traces are stored"
        )
parser.add_argument(
        "-d", "--dshares", default=2, type=int, help="number of shares"
        )
parser.add_argument(
        "-i", "--begin", default=0, type=int, help="initial index for the traces"
        )
parser.add_argument(
        "-e", "--end", default=4000, type=int, help="final index for the traces"
        )
args = parser.parse_args()

#########################
########################
print("# -----------")
print("# ----profile")
print("# -----------")
snr = None
D = args.dshares
for f in tqdm(range(args.Nfiles),desc="profiling files"):
    #### Loading data
    fname = args.fname+"_%d.npz"%(f)
    dic = np.load(fname, allow_pickle=True)
    nonce = np.frombuffer(np.array(dic["nonce"]).tobytes(),dtype=np.uint32).reshape((100000,4)).T
    k = np.frombuffer(np.array(dic["umsk_keys"]).tobytes(),dtype=np.uint32).reshape((100000,4)).T

    ### Computing and Shifting the state
    state = np.bitwise_xor(nonce,k)
    state = np.array(sbox_layer(state))
    state = state.T

    traces = dic["traces"][:,args.begin:args.end]
    s = (state[:,0] & 0x3) 
    s += (state[:,1] & 0x3) << 2
    s += (state[:,2] & 0x3) << 4
    s += (state[:,3] & 0x3) << 6

    if f == 0:
        snr = SNROrder(Nc=256,Ns=len(traces[0,:]),Np=1,D=D)

    snr.fit_u(traces,s.reshape(1,100000),nchunks=10)
    SM,u,std,_ = snr.get_sm(D=D)

    fname = args.outname+'.npz'
    np.savez(fname,SM=SM,u=u,s=std,snr=snr._SNR,mean = np.mean(traces,axis=0))

## Profiling is done here. 
## Next, we just do a demonstration for ProbeCircuit.
print("###############")
print("## simulating the circuit")
print("###############")
ram_in = np.frombuffer(dic["ram_in"][0,:].tobytes(),dtype=np.uint32)
nonce = np.frombuffer(dic["nonce"][0],dtype=np.uint32)
msk_key = np.frombuffer(dic["msk_keys"][0],dtype=np.uint32)
umsk_key = np.frombuffer(dic["umsk_keys"][0],dtype=np.uint32)

print("# shared key:")
print(msk_key)

print("# unshared key:")
print(sharing_inputs(msk_key,D=D)[0])

print("# sbox input:")
sbox_input = nonce ^ umsk_key
print(sbox_input)

# define cycles to probes 
cycles = np.arange(100).astype(np.uint32)

# initialize the probing object
probecircuit = ProbeCircuit(D=D,cycles=cycles,LIBSIM=args.simlib)

# obtain the register value at every probed cycles
reg_value = probecircuit.sim_run_probe(ram_in).reshape((len(cycles),D*4))

# change the reprensentation of the shares within the reg_value to fit the Python one at cycles
reg_value = rev_wire(reg_value,D=D)

# get the unmasked value of the register
reg_value_um = sharing_inputs(reg_value,D=D)

# get the cycle of interest for sbox_input
cycles_interest,_ = np.where(sbox_input[0] == reg_value_um)

print("# Sbox input is manipulated on cycle %d"%(cycles_interest[0]))
