# MIT License
#
# Copyright (c) 2020 UCLouvain
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from interface.spook import sbox_layer
import numpy as np
import argparse
from tqdm import tqdm
import pickle
import os
#########################
#########################
# Parsing args
#########################
parser = argparse.ArgumentParser(description="Attack phase of MCP-DPA",
        formatter_class=argparse.RawTextHelpFormatter
        )
parser.add_argument(
        "-t", "--tname", default="templates", help="file name where templates are saved"
)
parser.add_argument(
    "-f", "--fname", default="traces", help="file name where traces are saved"
)
parser.add_argument(
    "-N", "--Nfiles", default=1, type=int, help="Amount of files in which the traces are stored"
        )

args = parser.parse_args()

# parameters
D = 3
shift = 15

# loading the template file
fname = args.tname+'.npz'
dic = np.load(fname,allow_pickle=True)
model = dic["model"]
POIs= dic["POIs"]

u_0_l = dic["u_0_l"]
u_1_l = dic["u_1_l"]
u_2_l = dic["u_2_l"]

std_0_l = dic["std_0_l"]
std_1_l = dic["std_1_l"]
std_2_l = dic["std_2_l"]

# init variable to compute correlation on the fly
x = np.zeros(256)
xx = np.zeros(256)
xy = np.zeros(256)
yy = 0
y  = 0
n = 0


print("# ------------")
print("# --- attack")
print("# ------------",flush=True)

# Go through all the available files and update the correlation
for f in os.listdir(args.fname):
    # load the file
    fname = os.path.join(args.fname, f)
    print("# file: %s"%(fname),flush=True)
    dic = np.load(fname, allow_pickle=True)

    # shift to have bits of interest as LSBs
    nonce = dic["nonces"].T >> shift

    # compute the centered product for each of the four 2-bit words
    prod = [None for _ in range(4)]
    for v in range(4):
        l_0 = dic["traces"][:,POIs[v][0]].astype(np.float64)
        l_1 = dic["traces"][:,POIs[v][1]].astype(np.float64)
        l_2 = dic["traces"][:,POIs[v][2]].astype(np.float64)

        prod[v] = (l_0 - u_0_l[v])/std_0_l[v]
        prod[v] *= (l_1 - u_1_l[v])/std_1_l[v]
        prod[v] *= (l_2 - u_2_l[v])/std_2_l[v]
    prod = np.array(prod)
    prod = np.sum(prod,axis=0)

    nt = len(prod)
    n += nt

    # build the model for every key guess
    expected_moment = np.zeros((256,len(prod)),dtype=np.float64)
    for g in range(256):
        #putting the key guess in a bitlice fashion
        guess_state = np.zeros(4,dtype=np.uint32)
        guess_state[0] = (g>>0) & 0x3
        guess_state[1] = (g>>2) & 0x3
        guess_state[2] = (g>>4) & 0x3
        guess_state[3] = (g>>6) & 0x3

        # computing Sbox output
        guess_state = np.bitwise_xor(nonce.T,guess_state)
        guess_state = guess_state.T
        guess_state = sbox_layer(guess_state)

        # get Sbox output as an 8-bit word
        sbox_out = (guess_state[0] & 0x3) << 0
        sbox_out |= (guess_state[1] & 0x3) << 2
        sbox_out |= (guess_state[2] & 0x3) << 4
        sbox_out |= (guess_state[3] & 0x3) << 6

        # set model for this key
        expected_moment[g,:] = model[sbox_out]

    ## update the correlation
    x += np.sum(expected_moment,axis=1)
    xx += np.sum(expected_moment**2,axis=1)
    xy += np.sum(expected_moment*prod,axis=1)
    yy += np.sum(prod**2)
    y += np.sum(prod)

num = ((n*xy) - x*y)
den = (n*xx - x**2) * (n*yy - y**2)
corr = num/np.sqrt(np.abs(den))
sorted_keys = np.flip(np.argsort(np.abs(corr)))
key_guess = sorted_keys[0]
print("# %d traces: guess is %x "%(n,key_guess),flush=True)

# Turn correlation into log proba. 
# This is an heuristic
corr = corr**2
corr /= np.sum(corr)
lcorr = np.log10(corr)

# Compute the mask for the target bits
mask = ((0x3 << 0) |(0x3 << 32) | (0x3 << 64) | (0x3 << 96))<<(1)
mask_15 = (0x80000001 << 0) | (0x80000001 << 32) | (0x80000001 << 64) | (0x80000001 <<96)
masks = [(mask<<i) for i in range(0,32,2)]
masks[15] = mask_15
scores  =[np.zeros(256) for i in range(0,32,2)]
scores[shift//2] = lcorr

# print maskes on stdout
for (mask, score) in zip(masks, scores):
    print('{:x}:{}'.format(mask, ','.join(f'{x:.04f}' for x in score)))
exit()
