# MIT License
#
# Copyright (c) 2020 UCLouvain
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from stella_zero.estimator.accumulator import *
from stella_zero.evaluation.snr import SNR,SNROrder
import numpy as np
import argparse
from tqdm import tqdm
from interface.spook_masked import *

#########################
#########################
# Parsing args
#########################
parser = argparse.ArgumentParser(description="Performs the profiling for MCP-DPA",
        formatter_class=argparse.RawTextHelpFormatter
        )

parser.add_argument(
    "-f", "--fname", default="traces", help="file name where traces are saved"
)
parser.add_argument(
        "-o", "--outname", default="templates_mcp_dpa", help="file name where templates are saved"
)
parser.add_argument(
    "-N", "--Nfiles", default=1, type=int, help="Amount of files in which the traces are stored"
        )

args = parser.parse_args()

########################
# Number of shares
D = 3
# Number of classes to compute the SNR, 2 bits
Nc = 4
# Shift applied
shift = 15

# storage for shares value
acc_v = None
snr = None

# first pass on data to get the point of interest
print("--------------")
print("---- profiling")
print("--------------")
for f in tqdm(range(args.Nfiles),desc="Computing SNR on the shares"):
    #### Loading data
    fname = args.fname+'_%d.npz'%(f)
    dic = np.load(fname, allow_pickle=True)
    npub = dic["nonces"]
    seed = dic["seeds"]
    key = dic["msk_keys"]

    # Get the value of the shares after the Sbox.
    # Nr -> round executed
    # step -> at which step to stop
    masked_state = clyde128_encrypt_masked(npub.T,
                    np.zeros(4,dtype=np.uint32),
                    key.T,
                    seed.T,Nr=1,step=0,D=D)
    masked_state = masked_state.T

    # shift masked state to get the bits of interest as LSBs
    masked_state  = masked_state>>shift
    traces = dic["traces"][:,:]

    nt,_ = traces.shape
    if f == 0:
        # Declare SNR object for all the D*4 shares
        snr = SNR(Nc=Nc,Ns=len(traces[0,:]),Np=D*4)

        # accumulator to store shares values
        acc_v = [Accumulator(l=nt*args.Nfiles,Ns=D,dtype=np.uint8) for v in
                range(4)]

    # fit the snr with fresh traces
    snr.fit_u(traces,masked_state[:,:],nchunks=10)

    # keep track of the shares values
    for v,acc in enumerate(acc_v):
        acc.fit(masked_state[v*D:(v+1)*D,:].T%Nc)

# Find the MAX SNR point for each shares
POIs = [[np.argmax(snr._SNR[(v*D)+d,:]) for d in range(D)] for v in range(4)]

# second pass to store only the traces at the given POI
for f in tqdm(range(args.Nfiles),desc="Second pass to save traces at POI"):
    #### Loading data
    fname = args.fname+'_%d.npz'%(f)
    dic = np.load(fname, allow_pickle=True)
    traces = dic["traces"][:,:]
    nt,ns = traces.shape

    if f == 0:
        # One accumulator per share
        acc_t = [[Accumulator(l=nt*args.Nfiles,Ns=1,dtype=np.int16) for _ in
                range(D)] for _ in range(4)]
    for v,acc in enumerate(acc_t):
        [a.fit(traces[:,POIs[v][d],None]) for d,a in enumerate(acc)]

# get unshared value at the output of the sbox
sbox_out = 0
for v,acc in enumerate(acc_v):
    shares_value = acc.get().T
    share_0_v = shares_value[0]
    share_1_v = shares_value[1]
    share_2_v = shares_value[2]
    sbox_out |= (share_2_v ^ share_1_v ^ share_0_v) << (v*2)


# Compute the centered product for each of the four 2-bit words
u_0_l = [None for _ in range(4)]
u_1_l = [None for _ in range(4)]
u_2_l = [None for _ in range(4)]
std_0_l = [None for _ in range(4)]
std_1_l = [None for _ in range(4)]
std_2_l = [None for _ in range(4)]

prod = [None for _ in range(4)]
for v in range(4):
    share_0_l = acc_t[v][0].get()[:,0].astype(np.float64)
    share_1_l = acc_t[v][1].get()[:,0].astype(np.float64)
    share_2_l = acc_t[v][2].get()[:,0].astype(np.float64)

    u_0_l[v] = np.mean(share_0_l)
    u_1_l[v] = np.mean(share_1_l)
    u_2_l[v] = np.mean(share_2_l)

    std_0_l[v] = np.std(share_0_l)
    std_1_l[v] = np.std(share_1_l)
    std_2_l[v] = np.std(share_2_l)

    prod[v] = (share_0_l-u_0_l[v]) / std_0_l[v]
    prod[v] *= (share_1_l-u_1_l[v]) / std_1_l[v]
    prod[v] *= (share_2_l-u_2_l[v]) / std_2_l[v]

# Sum over the 4 words
prod = np.array(prod)
prod = np.sum(prod,axis=0)
model = np.zeros(256)

# compute the model for each possible sensible value
for i in range(256):
    model[i] = np.mean(prod[np.where(sbox_out==i)])

# save the template
fname = args.outname+'.npz'
np.savez(fname,u_0_l=u_0_l,u_1_l=u_1_l,u_2_l=u_2_l,
            std_0_l=std_0_l,std_1_l=std_1_l,std_2_l=std_2_l,
            POIs=POIs,model=model,allow_pickle=True)
